// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "windows.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// German (Austria) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_DEA)
LANGUAGE LANG_GERMAN, SUBLANG_GERMAN_AUSTRIAN
#pragma code_page(1252)

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include ""windows.h""\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON1               ICON                    ".\\img\\icons\\nomacs.ico"

IDI_ICON2               ICON                    ".\\img\\icons\\nomacs.ico"

IDI_ICON3               ICON                    ".\\img\\icons\\file-icon-jpg.ico"

IDI_ICON4               ICON                    ".\\img\\icons\\file-icon-animated.ico"

IDI_ICON5               ICON                    ".\\img\\icons\\file-icon-png.ico"

IDI_ICON6               ICON                    ".\\img\\icons\\file-icon-png-like.ico"

IDI_ICON7               ICON                    ".\\img\\icons\\file-icon-raw.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 3,8,0,1170
 PRODUCTVERSION 3,8,0,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "TU Wien"
            VALUE "FileDescription", "nomacs - Image Lounge"
            VALUE "FileVersion", "3.8.0.1170"
            VALUE "InternalName", "nomacs.exe"
            VALUE "LegalCopyright", "(C) 2011-2017, Markus Diem, Stefan Fiel, Florian Kleber"
            VALUE "OriginalFilename", "nomacs.exe"
            VALUE "ProductName", "nomacs"
            VALUE "ProductVersion", "3.8.0.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // German (Austria) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

